import click
from csv import DictReader
from .database import db_query
from .api_wrapper import request_data, tenb_connection
import csv
import datetime
import time, pprint


tio = tenb_connection()


def grab_hop_count(uuid):
    # grab the output of 10287 - Trace Route
    hop_count_data = db_query("select output from vulns where asset_uuid='{}' and plugin_id='10287';".format(uuid))

    # Send the raw data back
    return hop_count_data


def parse_19506_from_file(filename, scanid, histid):
    # Set Total Vars
    total_assets_scanned_list = []
    reported_scan_duration = None
    total_reported_scan_duration = None

    # Let's get the scan time with scan processing
    scan_history = request_data("GET", "/scans/{}/history".format(scanid))

    for hist in scan_history['history']:
        if not hist['is_archived']:
            if str(hist['id']) == str(histid):
                reported_scan_start = hist['time_start']
                reported_scan_end = hist['time_end']
                total_reported_scan_duration = reported_scan_end - reported_scan_start

                # Simple Delta from Tenable.io reported Scan times
                reported_scan_duration = str(datetime.timedelta(seconds=total_reported_scan_duration))

    # Open the file and parse the 19506 plugin
    with open(filename) as fobj:
        scan_name = "No Scan"
        scan_policy = "No Policy"
        scanner_ip = "No Scanner"
        scanner_list = []
        max_hosts = " "
        max_checks = ""
        asset_average_scantime = 0
        start_scan_timestamp_list = []
        timestamp_plus_duration_list = []
        assets_skipped = 0

        plugin_dict = {}

        for row in DictReader(fobj):
            plugin_output = row['Plugin Output']
            asset_uuid = row['Asset UUID']

            # split the output by return
            parsed_output = plugin_output.split("\n")

            # put everything but warnings into a dictionary
            for info_line in parsed_output:

                try:
                    new_split = info_line.split(" : ")
                    plugin_dict[new_split[0]] = new_split[1]

                except:
                    # Skip warnings
                    pass

            try:
                # Split at the colon to grab the numerical value
                intial_seconds = plugin_dict['Scan duration']

                # For an unknown reason, the scanner will print unknown for some assets leaving no way to calculate the time.
                if intial_seconds != 'unknown':

                    try:
                        # split to remove "secs"
                        number = intial_seconds.split(" ")

                        # grab the number for our minute calculation
                        final_number = number[0]

                        # Numerical value in seconds parsed from the plugin
                        seconds = int(final_number)

                        # Grab data pair and split it at the colon and grab the values
                        scan_name = plugin_dict['Scan name']
                        scan_policy = plugin_dict['Scan policy used']
                        scanner_ip = plugin_dict['Scanner IP']
                        # Enumerate all scanners for per/scanner stats
                        if scanner_ip not in scanner_list:
                            scanner_list.append(scanner_ip)
                        max_hosts = plugin_dict['Max hosts']
                        max_checks = plugin_dict['Max checks']
                        # Grabbing the start time from the plugin
                        scan_time = plugin_dict['Scan Start Date']

                        # Some timezone abbreviations are not parseable with strptime.
                        # removing the timezone for those we can't parse
                        try:
                            # Set the pattern to convert into epoch
                            pattern = '%Y/%m/%d %H:%M %z'
                            # Convert to Epoch
                            plugin_scan_time_epoch = int(time.mktime(time.strptime(scan_time, pattern)))

                        except ValueError:

                            # timezone couldn't be used. lets remove it and calculate what we can
                            pattern = '%Y/%m/%d %H:%M'
                            # Split the time to remove the timezone 2-6 chars
                            time_less_timezone = scan_time.split(" ")
                            # merge the data and time for calculation
                            new_time = "{} {}".format(time_less_timezone[0], time_less_timezone[1])
                            # Convert to Epoch
                            plugin_scan_time_epoch = int(time.mktime(time.strptime(new_time, pattern)))

                        # Export time pattern
                        new_pattern = '%Y-%m-%dT%H:%M:%S'

                        # Add to list to calculate the first asset scanned
                        start_scan_timestamp_list.append(plugin_scan_time_epoch)

                        # Add epoch and seconds to get end time.  This will be used to find the last asset scanned
                        timestamp_plus_duration_list.append(plugin_scan_time_epoch + seconds)

                        pltfm_start = row['Host Start'].split(".")[0]
                        pltfm_end = row['Host End'].split(".")[0]

                        new_start = time.mktime(time.strptime(pltfm_start, new_pattern))
                        new_end = time.mktime(time.strptime(pltfm_end, new_pattern))

                        # Platform End time in epoch minus the start time epoch
                        total_duration = new_end - new_start

                        # Total duration minus the length of the scan
                        indexing_time = total_duration - seconds

                        # All assets and 19506 seconds in a tuple
                        total_assets_scanned_list.append((asset_uuid, row['Host Start'], scan_time, seconds, indexing_time, total_duration, row['Host End'], row['IP Address']))

                        # pprint.pprint(total_assets_scanned_list)

                    except IndexError:
                        # This error occurs when an old scanner is used.
                        # the 19506 plugin filled with an error indicating the need for an upgrade
                        assets_skipped += 1
                        pass
                else:
                    # Print plugin output to identify an unknown plugin structure/response
                    click.echo(plugin_output)
            except KeyError as E:
                # If there is no scan duration, skip the asset.
                assets_skipped += 1
                pass
        # calculate the total for AVG calc
        total = 0
        index_total = 0
        for assets_scanned in total_assets_scanned_list:
            total += assets_scanned[3]
            index_total += assets_scanned[4]

        try:
            # Avg scan time per asset
            asset_average_scantime = total / len(total_assets_scanned_list)

            asset_average_indextime = index_total / len(total_assets_scanned_list)

            # total Scan time is: the oldest (scan_date in 19506 + scan duration in 19506) minus the earliest scan date in 19506
            total_calculated_scan_duration = (max(timestamp_plus_duration_list) - min(start_scan_timestamp_list))

            # Processing time therefore is the reported duration - the total scan duration
            processing = total_reported_scan_duration - total_calculated_scan_duration

            longest_scanned = max(total_assets_scanned_list, key=lambda uuid: uuid[3])
            shortest_scanned = min(total_assets_scanned_list, key=lambda uuid: uuid[3])

            longest_index = max(total_assets_scanned_list, key=lambda uuid: uuid[4])
            shortest_index = min(total_assets_scanned_list, key=lambda uuid: uuid[4])

            click.echo("*" * 100)
            click.echo("This data is derived from the 19506 plugins found in the {}"
                       "\nThe start time stamp(s) and the duration(s) were used to calculate the total scan time(s)\n"
                       "For manual examination of the details please check out the the file: {}-parsing.csv".format(filename, scanid))
            click.echo("*" * 100)

            click.echo("\nScan Name: {}".format(scan_name))
            click.echo("Scanner Policy: {}".format(scan_policy))

            # We can't count cloud scanners; so let's not try
            if "tenable.io Scanner" in scanner_list:
                click.echo("\nScanned Using Tenable Cloud Scanners".format(len(scanner_list)))
            else:
                click.echo("\nTotal Scanners Used: {}".format(len(scanner_list)))

                for scanner in scanner_list:
                    click.echo("Scanner IP: {}".format(scanner))

            click.echo("\nMax Hosts: {}".format(max_hosts))
            click.echo("Max Checks: {}\n".format(max_checks))
            click.echo("{} {}".format("Total Assets scanned: ", len(total_assets_scanned_list)))
            click.echo("{} {}\n".format("Number of Assets skipped", assets_skipped))

            click.echo("{:40} {:10} {}".format("Scan Stats", "H: M: S", "Asset"))
            click.echo("-" * 60)
            click.echo("{:40} {:10}".format("Total Reported Scan Duration:", reported_scan_duration))
            click.echo("{:40} {:10}".format("Total T.io Processing Time:", str(datetime.timedelta(seconds=processing))))
            click.echo("{:40} {:10}\n".format("Actual Scan Duration: ", str(datetime.timedelta(seconds=total_calculated_scan_duration))))
            click.echo("Asset Stats")
            click.echo("-" * 40)
            click.echo("{:40} {:10} {}".format("Longest Asset Duration: ", str(datetime.timedelta(seconds=longest_scanned[3])), longest_scanned[0]))
            click.echo("{:40} {:10} {}".format("Shortest Asset Duration: ", str(datetime.timedelta(seconds=shortest_scanned[3])), shortest_scanned[0]))
            click.echo("{:40} {:10} {}".format("Longest Asset Index: ", str(datetime.timedelta(seconds=longest_index[4])), longest_index[0]))
            click.echo("{:40} {:10} {}\n".format("Shortest Asset Index: ", str(datetime.timedelta(seconds=shortest_index[4])), shortest_index[0]))
            click.echo("Averages Stats")
            click.echo("-" * 40)
            click.echo("{:40} {}".format("Average scan duration per Asset:", str(datetime.timedelta(seconds=asset_average_scantime))))
            click.echo("{:40} {}\n".format("Average processing per Asset:", str(datetime.timedelta(seconds=asset_average_indextime))))

            click.echo("\nFirst asset scanned started at: {}".format(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(min(start_scan_timestamp_list)))))
            click.echo("Last asset finished scanning at: {}\n\n".format(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(max(timestamp_plus_duration_list)))))

        except ZeroDivisionError:
            click.echo("\nScan history had No data.\n")

    with open("13-parsing.csv", mode='w', encoding='utf-8', newline="") as csv_file:
        agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')
        header = ["UUID", "Platform_Start", "19506_scantime", "19506_duration","Indexing duration", "Total_duration",  "Platform_end", "IP Address"]

        agent_writer.writerow(header)
        # Loop through each asset
        for assets in total_assets_scanned_list:
            agent_writer.writerow(assets)


def get_last_history_id(scanid):
    data = request_data("GET", "/scans/{}/history".format(scanid))
    for scans in data['history']:
        if not scans['is_archived']:
            if scans['status'] == 'completed':
                return scans['id']
            else:
                click.echo("\nNo completed Scan to evaluate\n")
                exit()
        else:
            click.echo("\nNo completed Scan to evaluate\n")
            exit()


def download_csv_by_plugin_id(scan_id, hist_id):
    filename = f'{scan_id}-report.csv'
    click.echo("\nDownloading scan {} details into a csv called {} for "
               "parsing and manual auditing\n".format(scan_id, filename))

    # Stream the report to disk
    with open(filename, 'wb') as fobj:
        tio.scans.export(scan_id, ('plugin.id', 'eq', '19506'),
                         format='csv', fobj=fobj, history_id=hist_id)
    # Now parse the data
    parse_19506_from_file(filename, scan_id, hist_id)


def evaluate_a_scan(scanid, histid):

    if scanid:
        if histid:
            # use the scan id and hist id to download and parse
            download_csv_by_plugin_id(scanid, histid)
        else:
            # grab the last usable histid
            historyid = get_last_history_id(scanid)
            # download and parse
            download_csv_by_plugin_id(scanid, historyid)
    else:
        # Since no scanid was provided we assume the user wants stats on all scans
        click.echo("*" * 100)
        click.echo("\nThis command uses the 19506 plugin data found in the navi.db\n"
                   "Run a navi update command to refresh the database.\n")
        click.echo("*" * 100)
        # Pull all 19506 Plugins from the DB
        plugin_data = db_query("select asset_uuid, output from vulns where plugin_id='19506';")

        # Set some dicts for organizing Data
        scan_policy_dict = {}
        scanner_dict = {}
        scan_name_dict = {}

        # Open a CSV for export
        with open('evaluate.csv', mode='w', encoding='utf-8', newline="") as csv_file:
            agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

            header_list = ["asset uuid", "Scan Name", "Scan Policy", "Scanner IP", "Scan Time", "Max Checks", "Max Hosts", "Minutes", "RTT", "Hop Count"]

            # Write the header to the csv
            agent_writer.writerow(header_list)

            # This function is used to parse the data
            # Getting the length of the Category and using it to get the average
            def average_by_policy(name, scan_info):
                # Print the Category to the Screen ( Scanner, Policy, Scan Name)
                click.echo("\n{:100s} {:25s} {:10}".format(name, "AVG Minutes Per/Asset", "Total Assets"))
                click.echo("-" * 20)

                # Cycle through each category
                for scan in scan_info.items():

                    # data is in a list [asset_uuid, mins] We need the length of the total mins found
                    length = len(scan[1])

                    # Reset the total per Category Item - Specific Scan ID, Scanner, Policy ID
                    total = 0

                    # Cycle through each asset record
                    for assets in scan[1].values():
                        # Gather a total
                        total = assets + total

                    # After calculating the total, lets get an average
                    average = total/length

                    # Print results to the screen
                    click.echo("\n{:100s} {:25d} {:10d}".format(scan[0], int(average), length))

                click.echo("-" * 150)

            # Loop through each plugin 19506 and Parse data from it
            for vulns in plugin_data:

                # Output is the second item in the tuple from the DB
                plugin_output = vulns[1]

                # split the output by return
                parsed_output = plugin_output.split("\n")

                # grab the length so we can grab the seconds
                plugin_length = len(parsed_output)

                # grab the scan duration- second to the last variable
                duration = parsed_output[plugin_length - 2]

                # Split at the colon to grab the numerical value
                seconds = duration.split(" : ")

                try:
                    # split to remove "secs"
                    number = seconds[1].split(" ")
                except:
                    # No plugin Data for this asset
                    pass

                # grab the number for our minute calculation
                final_number = number[0]

                if final_number != 'unknown':
                    try:
                        # convert seconds into minutes
                        minutes = int(final_number) / 60

                        # Grab data pair and split it at the colon and grab the values
                        scan_name = parsed_output[9].split(" : ")[1]
                        scan_policy = parsed_output[10].split(" : ")[1]
                        scanner_ip = parsed_output[11].split(" : ")[1]
                        scan_time = parsed_output[plugin_length - 3].split(" : ")[1]
                        max_hosts = parsed_output[plugin_length- 8].split(" : ")[1]
                        max_checks = parsed_output[plugin_length - 7].split(" : ")[1]

                        if "no" not in parsed_output[14].split(" : ")[1]:
                            rtt = parsed_output[14].split(" : ")[1]
                        else:
                            rtt = parsed_output[12].split(" : ")[1]

                        try:
                            # Grab the last line in the Trace route Plugin output
                            # Split on the space and grab the numerical value.
                            hopcount = grab_hop_count(vulns[0])[0][-1].split(" ")[-1]
                        except IndexError:
                            hopcount = "Unknown"

                        parsed_data_organized = [vulns[0], scan_name, scan_policy, scanner_ip, scan_time, max_checks, max_hosts, minutes, rtt, hopcount]

                        agent_writer.writerow(parsed_data_organized)

                        # Organize Data by Scan Policy
                        # If the category is not in the new dict, add it; else update it.
                        if scan_policy not in scan_policy_dict:
                            scan_policy_dict[scan_policy] = {vulns[0]: minutes}
                        else:
                            scan_policy_dict[scan_policy].update({vulns[0]: minutes})

                        if scanner_ip not in scanner_dict:
                            scanner_dict[scanner_ip] = {vulns[0]: minutes}
                        else:
                            scanner_dict[scanner_ip].update({vulns[0]: minutes})

                        if scan_name not in scan_name_dict:
                            scan_name_dict[scan_name] = {vulns[0]: minutes}
                        else:
                            scan_name_dict[scan_name].update({vulns[0]: minutes})
                    except IndexError:
                        # This error occurs when an old scanner is used.
                        # the 19506 plugin filled with an error indicating the need for an upgrade
                        pass

            average_by_policy("Policies", scan_policy_dict)

            average_by_policy("Scanners", scanner_dict)

            average_by_policy("Scan Name", scan_name_dict)




from .api_wrapper import tenb_connection, request_data
import click

tio = tenb_connection()


@click.command(help="Create Tag rules in Tenable.io")
@click.option('--c', default='', help="Create a Tag with the following Category name")
@click.option('--v', default='', help="Create a Tag Value; requires --c and Category Name or UUID")
@click.option('--d', default='This Tag was created/updated by navi', help="Description for your Tag")
@click.option('--filter', default='', help="Filter used to tag assets")
@click.option('--action', default='', help="Type of operator")
@click.option('--value', default='', help="Filter value")
@click.option('--multi', default='', help="A well crafted pytenable list of tuples")
@click.option('-any', is_flag=True, help="Change the default from 'and' to 'or")
def tagrule(c, v, filter, action, value, d, multi, any):
    if c == '':
        click.echo("Category is required.  Please use the --c command")
        exit()

    if v == '':
        click.echo("Value is required. Please use the --v command")
        exit()

    if multi == '' and filter == '':
        click.echo("\nYou need to select an individual 'filter' or 'multi' for multiple filters\n")
        exit()

    if multi:
        click.echo("creating a new tag with the following list {}".format(multi))
        filter_list = []
        for os in eval(multi):
            temp_dict = {"field": os[0], "operator": os[1], "value": os[2]}
            filter_list.append(temp_dict)
            # issue with tenable.io API - Or turns in to a string value seperated by a comma
            # issue with pytenable - appends " when you add * so *centos* becomes "*centos*"
        if any:
            tio.tags.create(c, v, filters=eval(multi), filter_type="or", description=d)
        else:
            tio.tags.create(c, v, filters=eval(multi), description=d)
        #payload = {"filters": {"asset": {"or": filter_list}}, "category_name": str(c), "value": str(v), "description": str(d)}
        import pprint
        #pprint.pprint(payload)
        #data = request_data('POST', '/tags/values', payload=payload)
        #print(data)

    if filter:
        if action:
            if value:
                rule_tuple = (filter, action, [value])
                tio.tags.create(c, v, filters=[rule_tuple], description=d)
            else:
                click.echo("You must have a value if you are going to use a filter")
            exit()
        else:
            click.echo("You must have an Action if you are going to use a filter")
            exit()



import click
import os
from .database import new_db_connection


def grab_keys():
    database = r"navi.db"
    conn = new_db_connection(database)
    with conn:
        cur = conn.cursor()
        try:
            cur.execute("SELECT * from keys;")
        except:
            click.echo("\nYou don't have any API keys!  Please enter your keys\n")
            exit()
        rows = cur.fetchall()
        for row in rows:
            access_key = row[0]
            secret_key = row[1]

    return access_key, secret_key


@click.group(help="Deploy a Navi front-end using Docker")
def deploy():
    pass


@deploy.command(help="Deploy Navi Tag Center using a Docker container: navigate to http://localhost:5000")
def tag_center():
    if click.confirm('This command downloads the packetchaos/tag-center docker container and runs it on port 5000 using the current navi database. Deploy?'):
        try:
            os.system("docker run -d -p 5000:5000 --mount type=bind,source=\"$(pwd)\",target=/usr/src/app/data packetchaos/tag-center")
        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Deploy Navi Was Reporter using a Docker container: navigate to http://localhost:5004")
@click.option("--days", default=60, help="Limit the amount of data being downloaded/reported")
def was_reporter(days):
    a, s = grab_keys()
    command = "docker run -d -p 5004:5004 -e \"access_key={}\" -e \"secret_key={}\" -e {} --mount type=bind,source=$(pwd),target=/usr/src/app/data packetchaos/navi_was_reports".format(a,s,days)
    if click.confirm('This command downloads the packetchaos/navi_was_reports docker container and runs it on port 5004 using the current navi database. Deploy?'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Deploy Navi Scantime Tagging solution")
def scantags():
    a, s = grab_keys()
    command = "docker run -d packetchaos/scantags {} {}".format(a,s)
    if click.confirm('This command downloads the packetchaos/scantags docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Deploy Navi Discovery then Vuln Scan solution")
@click.option('--trigger', default=None, help="The Scan policy ID you want to use as the the Trigger Scan, or the first scan in the chain.")
@click.option('--fire', default=None, help="The scan policy ID you want to use for your Vuln Scan")
@click.option('--targets', default=None, help='The subnet(s) you want to run the discovery scan on.')
def discoverythenvulnscan(trigger, fire, targets):
    a, s = grab_keys()
    command = "docker run -d packetchaos/discovery_then_vulnscan {} {} {} {} {}".format(a, s, trigger, fire, targets)
    if click.confirm('This command downloads the packetchaos/discoverythenvulnscan docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Deploy Navi Critical Tags Docker solution")
def critical_tags():
    a, s = grab_keys()
    command = "docker run -d packetchaos/critical_tags {} {}".format(a,s)
    if click.confirm('This command downloads the packetchaos/critical_tags docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Tag each asset by the agent group membership")
def agent_group_tags():
    a, s = grab_keys()
    command = "docker run -d packetchaos/agent_group_tags {} {}".format(a,s)
    if click.confirm('This command downloads the packetchaos/agent_group_tags docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Tag each asset by the ports found open")
def port_tagging():
    a, s = grab_keys()
    command = "docker run -d packetchaos/port_tagging {} {}".format(a,s)
    if click.confirm('This command downloads the packetchaos/port_tagging docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")


@deploy.command(help="Deploy the All tags solution.  Deploy all tags from all the navi services")
def all_tags():
    a, s = grab_keys()
    command = "docker run -d packetchaos/all_tags {} {}".format(a,s)
    if click.confirm('This command downloads the packetchaos/all_tags docker container and runs it.  This will run as a service and will be destroyed after the all assets are tagged.'):
        try:
            os.system(command)

        except os.error:
            click.echo("You might not have Docker installed")



import re
import random
import string
import pandas as pd
import numpy as np
from collections import defaultdict
from typing import Optional

PASSWORD_LENGTH = 20


class Excel:
    def __init__(self, file_path: str, sheet_names: Optional[list]=None) -> None:
        self.excel = pd.ExcelFile(file_path)
        self._file_path = file_path
        # include all sheets or just ones that were specified in sheet_names
        if sheet_names is not None:
            self.sheet_names = [name for name in self.excel.sheet_names if name in sheet_names]
        else:
            self.sheet_names = self.excel.sheet_names

        self.sheets = {name: self.parse_sheet(name) for name in self.sheet_names}

    def get_sheet_action(self, sheet_name, action):
        return [record for record in self.records[sheet_name] if record['action'] == 'action']

    def parse_sheet(self, sheet_name: str):
        df = self.excel.parse(sheet_name)
        df = df.applymap(strip_whitespace)
        records = df.replace(np.nan, None).to_dict('records')
        return records

    def get_records(self, action='create', asset_tag_filters=None):
        return post_process_sheets(self.sheets, asset_tag_filters=asset_tag_filters, action=action)


def strip_whitespace(value):
    if isinstance(value, str):
        value = re.sub('\n+', ',', value.strip())
        value = re.sub('[ ]*,[ ]*', ',', value.strip())
    return value

def generate_password(length=PASSWORD_LENGTH):
    special_chars = '`~!@#$%^&*()-_+='
    password_chars = string.digits + string.ascii_letters + special_chars
    return "".join(random.choices(password_chars, k=length))

def insert_password(user):
    user['password'] = generate_password()
    return user

def username_to_lower(user):
    user['username'] = user['username'].lower()
    return user

def str_to_api_name(name):
    if name.startswith('tags') or name.endswith('tags'):
        name = 'tags'
    return name


def parse_filter_name(column_name: str, asset_tag_filters: dict) -> tuple[dict, str]:
    '''split value into (filter_name, operator)

    The filter column name will be either:
        - the name of a filter, i.e. ipv4
        - OR the <name><' ' or '_'><operator

    '''
    # default to equal when the value is a filter_name with out the operator
    operator = 'eq'
    filter_name = column_name

    # get the dictionary of valid filter names and operators
    tag_filter = asset_tag_filters.get(column_name)

    if tag_filter is None:
        # see if there is an operator appended to the filter_name
        match = re.match('(?P<filter_name>\w+)[ _](?P<operator>\w+)', column_name)
        if match is None:
            raise ValueError(f'[{column_name}]: bad format')

        filter_name, operator = match.groups()
        tag_filter = asset_tag_filters.get(filter_name)

        if tag_filter is None:
            raise KeyError(f'[{filter_name}]: filter name not found')

        if operator not in tag_filter['operators']:
            raise KeyError(f'[{operator}]: not in {tag_filter["operators"]}')

    return filter_name, operator


def build_filters(columns, asset_tag_filters: dict):
    '''Build the filter statement based on multiple columns.'''
    header_fields=('category', 'value', 'filter_type')
    # filter items are all k, v pairs execept
    filter_values = {k: v for k, v in columns.items() if k not in header_fields}
    filters = []
    for field_name, field_value in filter_values.items():
        filter_name, operator = parse_filter_name(field_name, asset_tag_filters)
        if field_value is None:
            continue

        field_value = re.sub('\n+', ',', field_value.strip())
        field_value = re.sub('[ ]*,[ ]*', ',', field_value)

        filters.append((filter_name, operator, field_value))


    # build a record with only the header file
    record = {k: v for k, v in columns.items() if k in header_fields}
    record['filters'] = filters

    return record


def process_groups_from_users(records):

    membership = defaultdict(list)
    group_commands = []

    for user in records:
        if 'groups' not in user['record']:
            continue

        # build group commands and membership index
        for group_name in user['record']['groups'].split(','):
            membership[group_name].append(user['record']['username'])

    # build group 'create'
    for group_name in membership:
        group_cmd = {
            'api_name': 'groups',
            'action': 'create',
            'record': {'name': group_name}
        }
        group_commands.append(group_cmd)

    # build group 'add_user' from membership info
    for group_name, members in membership.items():
        usernames = sorted([u for u in members])
        for name in usernames:
            group_cmd = {
                'api_name': 'groups',
                'action': 'add_user',
                'record': {
                    'group_id': None,
                    'user_id': None,
                    'group_name': group_name,
                    'username': name
                }
            }
            group_commands.append(group_cmd)

    return group_commands


def post_process_sheets(sheets: dict[str,list], asset_tag_filters: dict = None, action: str = None):
    '''Process sheets to do things build filters from columns and identify groups within user records

    Args:
        sheets: lists of records for each sheet, indexed by sheet name
        asset_tag_filters: valid tag filters from tio.filters.asset_tag_filters()

    '''

    for name, records in sheets.items():
        # we may have to convert the sheetname, i.e. tags_ipv4 or ipv4_tags to tags
        api_name = str_to_api_name(name)

        # generate a password for each user record
        if api_name == 'users':
            records = map(insert_password, records)

        # build filter parameters from columns
        if api_name == 'tags' and asset_tag_filters is not None:
            # records = map(build_filters, records)
            records = [build_filters(record, asset_tag_filters) for record in records]

        def encapsulate(r):
            return {'api_name': api_name, 'action': r.get('action', action), 'record': r}

        sheets[name] = [encapsulate(r) for r in records]

    # expand 'group create' and 'group add user' for groups defined in the user records
    if 'users' in sheets:
        groups = process_groups_from_users(sheets['users'])
        if 'groups' in sheets:
            sheets['groups'].append(groups)
        else:
            sheets['groups'] = groups

    # combine multiple sheets of tags in to a single 'tags' entry
    tag_sheets = {k: v for k, v in sheets.items() if k.startswith('tags') or k.endswith('tags')}
    tag_records = [record for records in tag_sheets.values() for record in records]

    # remove identified tag_sheets from sheets, add combined lists of records to sheets['tags']
    for name in list(tag_sheets):
        del sheets[name]
    sheets['tags'] = tag_records

    return sheets


def main():
    ws = Excel('tio-config.xlsx', sheet_names='users')
    _records = ws.get_records()

if __name__ == '__main__':
    main()



import click
from .api_wrapper import tenb_connection

tio = tenb_connection()


@click.command(help="Create a Scanner Group")
@click.option("--name", required=True, help="Name of Scanner Group")
def sgroup(name):
    tio.scanner_groups.create(name)



from .dbconfig import create_apps_table, new_db_connection, create_plugins_table
from .database import insert_apps, insert_plugins, drop_tables
from .api_wrapper import request_data
import click
import time
import datetime

def plugin_parser(plugin_output):
    tech_list = []
    # Split the plugin information on '-'
    plugin_tuple = plugin_output.split('-')
    # Ignore the item in the tuple and add all others to a list
    for x in range(len(plugin_tuple) - 1):
        tech_list.append(str(plugin_tuple[x + 1]))
    return tech_list


def vuln_counter(plugin_id, scan_uuid):
    database = r"navi.db"
    conn = new_db_connection(database)
    with conn:
        cur = conn.cursor()
        cur.execute("SELECT count(*) from plugins where plugin_id =='{}' and scan_uuid=='{}';".format(plugin_id, scan_uuid))

        plugin_data = cur.fetchall()

        return plugin_data[0][0]


def get_was_stats(scan_id):
    params = {"limit": "200", "offset": "0"}
    was_data = request_data("POST", "/was/v2/scans/{}/vulnerabilities/search".format(scan_id), params=params)
    stat_dict = {}

    for finding in was_data['items']:
        if str(finding['plugin_id']) == '98000':

            scan_meta_data = finding['details']['output']

            new_data = str(scan_meta_data).split()

            stat_dict['eng_version'] = "0"#new_data[2]
            stat_dict['start_time'] = "0"#"{} {} {}".format(new_data[11], new_data[12],new_data[13])
            stat_dict['duration'] = "0"#new_data[15]

            stat_dict['requests_made'] = "0"#new_data[17]
            stat_dict['crawler_requests'] = "0"#new_data[20]
            stat_dict['requests_per_sec'] = "0"#new_data[22]
            stat_dict['mean_response_time'] = "0"#new_data[26]

            stat_dict['data_target'] = "0"#"{} {}".format(new_data[33], new_data[34])
            stat_dict['target_to_data'] = "0"#"{} {}".format(new_data[39], new_data[40])

            stat_dict['network_timeouts'] = "0"#new_data[45]
            stat_dict['browser_timeouts'] = "0"#new_data[48]
            stat_dict['browser_respawns'] = "0"#new_data[51]

            return stat_dict


def download_data(uuid, asset):
    database = r"navi.db"
    app_conn = new_db_connection(database)
    app_conn.execute('pragma journal_mode=wal;')
    with app_conn:
        apps_table_list = []
        report = request_data('GET', '/was/v2/scans/{}/report'.format(uuid))
        #scan_metadata = get_was_stats(uuid)

        config_id = report['config']['config_id']

        # Ignore all scans that have not completed
        if report['scan']['status'] == 'completed':
            scan_name = report['config']['name']

            scan_completed_time = report['scan']['finalized_at']
            try:
                requests_made = 0#scan_metadata['requests_made']
            except KeyError:
                requests_made = 0

            try:
                pages_crawled = 0#scan_metadata['crawler_requests']
            except KeyError:
                pages_crawled = 0

            critical = []
            high = []
            medium = []
            low = []
            info = []
            critical_summary = []
            high_summary = []
            medium_summary = []
            low_summary = []
            info_summary = []
            tech_list = ['Nothing Found']
            owasp_list = []
            owasp_dict = {}
            try:
                notes = report['config']['notes']
            except KeyError:
                notes = "No Scan Notes"

            try:
                target = report['scan']['target']
            except KeyError:
                target = report['config']['settings']['target']

            # Count for-loop
            plugin_list = []

            for finding in report['findings']:
                plugin_list.append(finding['plugin_id'])
                for xref in finding['xrefs']:
                    # Grab multiples values here
                    if xref['xref_name'] == 'OWASP':
                        if '2021' in xref['xref_value']:
                            owasp_clean = str(xref['xref_value']).split('-')[1]
                            owasp_list.append(owasp_clean)

            def occurances(number, number_list):
                return number_list.count(number)

            for owasp in range(1, 11):
                owasp_dict["A{}".format(owasp)] = occurances("A{}".format(owasp), owasp_list)

            for finding in report['findings']:
                finding_list = []
                risk = finding['risk_factor']
                plugin_id = finding['plugin_id']
                plugin_name = finding['name']
                family = finding['family']
                cves = finding['cves']
                description = finding['description']
                output = finding['output']
                owasp = finding['owasp']
                payload = finding['payload']
                plugin_mod_date = finding['plugin_modification_date']
                plugin_pub_date = finding['plugin_publication_date']
                proof = finding['proof']
                request_headers = finding['request_headers']
                response_headers = finding['response_headers']
                solution = finding['solution']
                url = finding['uri']
                xrefs = finding['xrefs']
                see_also = finding['see_also']

                finding_list.append(str(uuid))
                finding_list.append(str(plugin_name))
                finding_list.append(str(cves))
                finding_list.append(str(description))
                finding_list.append(str(family))
                finding_list.append(str(output))
                finding_list.append(str(owasp))
                finding_list.append(str(payload))
                finding_list.append(str(plugin_id))
                finding_list.append(str(plugin_mod_date))
                finding_list.append(str(plugin_pub_date))
                finding_list.append(str(proof))
                finding_list.append(str(request_headers))
                finding_list.append(str(response_headers))
                finding_list.append(str(risk))
                finding_list.append(str(solution))
                finding_list.append(str(url))
                finding_list.append(str(xrefs))
                finding_list.append(str(see_also))

                insert_plugins(app_conn, finding_list)

                if str(plugin_id) == '98059':
                    tech_list = plugin_parser(finding['output'])

                vuln_count = occurances(finding['plugin_id'], plugin_list)
                vuln_list = [risk, plugin_id, plugin_name, family, vuln_count]
                if risk == 'high':
                    high.append(plugin_id)
                    if vuln_list not in high_summary:
                        high_summary.append(vuln_list)
                elif risk == 'medium':
                    medium.append(plugin_id)
                    if vuln_list not in medium_summary:
                        medium_summary.append(vuln_list)
                elif risk == 'low':
                    low.append(plugin_id)
                    if vuln_list not in low_summary:
                        low_summary.append(vuln_list)
                elif risk == 'critical':
                    critical.append(plugin_id)
                    if vuln_list not in critical_summary:
                        critical_summary.append(vuln_list)
                else:
                    info.append(plugin_id)
                    if vuln_list not in info_summary:
                        info_summary.append(vuln_list)

            apps_table_list.append(scan_name)
            apps_table_list.append(uuid)
            apps_table_list.append(target)
            apps_table_list.append(scan_completed_time)
            apps_table_list.append(pages_crawled)
            apps_table_list.append(requests_made)
            apps_table_list.append(len(critical))
            apps_table_list.append(len(high))
            apps_table_list.append(len(medium))
            apps_table_list.append(len(low))
            apps_table_list.append(len(info))
            apps_table_list.append(str(owasp_dict))
            apps_table_list.append(str(tech_list))
            apps_table_list.append(config_id)
            apps_table_list.append(str(notes))
            apps_table_list.append(str(asset))

            insert_apps(app_conn, apps_table_list)

    return


def grab_scans(days):
    click.echo("\nDownloading all Completed Scans for the last {} days.\n"
               "This will take some time.\n".format(days))
    database = r"navi.db"
    app_conn = new_db_connection(database)
    app_conn.execute('pragma journal_mode=wal;')

    drop_tables(app_conn, 'apps')
    drop_tables(app_conn, 'plugins')

    create_apps_table()

    create_plugins_table()
    data = request_data('POST', '/was/v2/configs/search?limit=200&offset=0')
    for configs in data['items']:
        config_id = configs['config_id']
        was_config_data = request_data("POST", "/was/v2/configs/{}/scans/search".format(config_id))
        # Ignore all scans that have not completed

        for scanids in was_config_data['items']:

            # check and skip parent records; parent records don't have an asset_uuid
            try:

                day = 86400
                new_limit = day * int(days)
                day_limit = time.time() - new_limit

                if scanids['status'] == 'completed' and scanids['template_name'] == 'scan':
                    try:
                        asset_uuid = scanids['asset_id']
                    except KeyError:
                        asset_uuid = "NO ID Found"

                    try:
                        uri = scanids['application_uri']
                    except KeyError:
                        uri = "NOPE"

                    if uri != "NOPE":
                        was_scan_id = scanids['scan_id']
                        finalized_at = scanids['finalized_at']
                        try:
                            epoch = datetime.datetime.strptime(finalized_at, "%Y-%m-%dT%H:%M:%S.%fZ").timestamp()

                            if epoch >= day_limit:

                                download_data(was_scan_id, asset_uuid)
                        except TypeError:
                            pass
                else:
                    pass
            except KeyError:
                pass



import click
from .api_wrapper import request_data, tenb_connection
import time
import uuid
import csv
import textwrap
from .database import db_query


tio = tenb_connection()


def web_app_scanners():
    click.echo("\nHere are the available scanners")
    click.echo("Remember, don't pick a Cloud scanner for an internal Site\n")
    scanners = request_data('GET', '/scanners')

    click.echo("{:30s} {}".format("Scanner Name", "Scanner UUID"))
    click.echo("-" * 50)
    for scanner in scanners['scanners']:
        if str(scanner['supports_webapp']) == 'True':
            click.echo("{:30s} {}".format(str(scanner['name']), str(scanner['id'])))
    click.echo()


def display_users():
    users = request_data('GET', '/users')
    click.echo("\n{:40s} {}".format("User Name", "User UUID"))
    click.echo("-" * 50)
    for user in users['users']:
        click.echo("{:40s} {}".format(str(user['user_name']), str(user['uuid'])))
    click.echo()


def create_was_scan(owner_id, temp_id, scanner_id, target, name):
    # use time to ensure UUID is correct
    uuid_offset = int(time.time())

    # create a new WAS UUID
    was_uuid = uuid.uuid1(uuid_offset)

    # Construct the payload
    payload = dict({
        "name": str(name),
        "owner_id": str(owner_id),
        "template_id": str(temp_id),
        "scanner_id": str(scanner_id),
        "target": str(target)
        }
    )

    request_data('PUT', '/was/v2/configs/' + str(was_uuid), payload=payload)
    return


@click.group(help="Interact with WAS V2 API")
def was():
    pass


@was.command(help="[STEP TWO] - Display Scan IDs for a given Scan Config")
@click.argument('config_uuid')
def scans(config_uuid):
    params = {"limit": "200", "offset": "0"}
    was_data = request_data("POST", "/was/v2/configs/{}/scans/search".format(config_uuid), params=params)

    click.echo("\n{:70s} {:15s} {:15s} {:40} {}".format("Target FQDN", "Audited URLS", "Found URLS", "Scan_ID", "Status"))
    click.echo("-" * 150)
    for app_data in was_data['items']:
        try:
            app_url = app_data['application_uri']
        except KeyError:
            app_url = "API issue occurred"

        app_scan_id = app_data['scan_id']
        try:
            app_audited_urls = app_data['metadata']['audited_urls']
        except TypeError:
            app_audited_urls = 0
        except KeyError:
            app_audited_urls = 0
        try:
            app_found_urls = app_data['metadata']['found_urls']
        except TypeError:
            app_found_urls = 0
        except KeyError:
            app_found_urls = 0
        try:
            app_status = app_data['status']
        except KeyError:
            app_status = "None"
        try:
            updated = app_data['updated_at']
        except KeyError:
            updated = "None"

        click.echo("{:70s} {:15s} {:15s} {:40} {}".format(textwrap.shorten(str(app_url), width=70), str(app_audited_urls), str(app_found_urls), str(app_scan_id), str(app_status), str(updated)))

    click.echo()


@was.command(help="Start a Web Application Scan")
@click.argument('scan_id')
def start(scan_id):
    click.echo("\n Your Scan is starting")
    request_data('POST', '/was/v2/configs/' + str(scan_id) + '/scans')


@was.command(help=" [STEP THREE] - Display Details for a Web Application Scan")
@click.argument('scan_uuid')
@click.option('--plugin', default='', help="Get Plugin output, response headers and request headers")
def details(scan_uuid, plugin):

    if plugin:
        plugin_data = db_query("select payload, output, request_headers, response_headers, description, solution, cves from plugins where scan_uuid='{}' and plugin_id='{}'".format(scan_uuid, plugin))

        click.echo("Description")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][4])
        click.echo()
        click.echo("Payload")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][0])
        click.echo()
        click.echo("Request Headers")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][2])
        click.echo()
        click.echo("Response Headers")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][3])
        click.echo()
        click.echo("Output")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][1])
        click.echo()
        click.echo("Solution")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][5])
        click.echo()
        click.echo("CVEs")
        click.echo("-" * 35)
        click.echo()
        click.echo(plugin_data[0][6])
        click.echo()

    else:
        detail_report_data = db_query("select * from plugins where scan_uuid='{}'".format(scan_uuid))
        detail_high = []
        detail_meduim = []
        detail_low = []
        click.echo("-" * 40)
        click.echo("\n{:10s} {:60s} {:10}".format("Plugin", "Plugin Name", "Severity"))
        click.echo("-" * 150)
        for detail_finding in detail_report_data:
            detail_plugin_id = detail_finding[8]
            detail_risk = detail_finding[14]
            plugin_name = detail_finding[1]

            if detail_risk == 'high':
                detail_high.append(detail_plugin_id)
            elif detail_risk == 'medium':
                detail_meduim.append(detail_plugin_id)
            elif detail_risk == 'low':
                detail_low.append(detail_plugin_id)

            click.echo("{:10s} {:60s} {:10} ".format(str(detail_plugin_id), textwrap.shorten(str(plugin_name), width=60),
                                                 str(detail_risk)))

        click.echo("\nSeverity Counts")
        click.echo("-" * 20)
        click.echo("High: {}".format(len(detail_high)))
        click.echo("Medium: {}".format(len(detail_meduim)))
        click.echo("Low: {}".format(len(detail_low)))
        click.echo()


@was.command(help="Scan a Web Application Target")
@click.argument('scan_target')
@click.option('--file', is_flag=True, help="File name of the CSV containing Web Apps for bulk scan creation")
def scan(scan_target, file):
    click.echo("\nChoose your Scan Template")
    click.echo("1.   Web App Overview")
    click.echo("2.   Web App Scan")
    click.echo("3.   WAS SSL SCAN")
    click.echo("4.   WAS Config Scan")
    option = input("Please enter option #.... ")
    if option == '1':
        template = "b223f18e-5a94-4e02-b560-77a4a8246cd3"
    elif option == '2':
        template = "112f3e7f-d83a-4bba-b2c8-df2d22e2fa4c"
    elif option == '3':
        template = "072f4d6b-1dd7-4049-b279-78a56d1c778e"
    elif option == '4':
        template = "3078d0c6-6e81-44de-b585-6921b69ff0ef"
    elif len(option) == 36:
        template = str(option)
    else:
        click.echo("Using Basic scan since you can't follow directions")
        template = "b223f18e-5a94-4e02-b560-77a4a8246cd3"

    # Display scanners
    web_app_scanners()

    # Capture Scanner selection
    scanner_id = input("What scanner do you want to scan with ?.... ")

    # Display Users UUID
    display_users()
    # Capture User UUID selection
    user_uuid = input("Select an Scan owner using the UUID ?.... ")

    scan_name = "navi Created Scan of : " + str(scan_target)

    if file:
        with open(scan_target, 'r', newline='') as csv_file:
            web_apps = csv.reader(csv_file)
            for apps in web_apps:
                for app in apps:
                    scan_name = "navi Created Scan of : " + str(app)
                    # strip for white spaces
                    application = app.strip()
                    create_was_scan(owner_id=user_uuid, scanner_id=scanner_id, name=scan_name, temp_id=template, target=application)
                    time.sleep(2)
    else:
        click.echo("\nCreating your scan now for site: " + str(scan_target))
        create_was_scan(owner_id=user_uuid, scanner_id=scanner_id, name=scan_name, temp_id=template, target=scan_target)


@was.command(help="[STEP ONE] - Display Web Application Configs")
def configs():
    # This doesn't hit the database to provide visibility in to all status' not just completed
    params = {"limit": "200", "offset": "0"}
    was_data = request_data("POST", "/was/v2/configs/search", params=params)

    click.echo("\n{:70s} {:40s} {:14s} {}".format("Config Name", "Config ID", "Status", "Last Update"))
    click.echo("-" * 150)
    for app_data in was_data['items']:
        try:
            app_name = app_data['name']
            app_scan_id = app_data['config_id']
            app_status = app_data['last_scan']['status']
            updated = app_data['updated_at']

            click.echo("{:70s} {:40s} {:14s} {}".format(textwrap.shorten(str(app_name), width=70), str(app_scan_id), str(app_status), str(updated)))
        except:
            pass
    click.echo()


@was.command(help="Display Statistics for Web Application Scan")
@click.argument('scan_id')
def stats(scan_id):
    params = {"limit": "200", "offset": "0"}
    was_data = request_data("POST", "/was/v2/scans/{}/vulnerabilities/search".format(scan_id), params=params)

    for finding in was_data['items']:
        if str(finding['plugin_id']) == '98000':

            scan_meta_data = finding['details']['output']
            print(scan_meta_data)

'''
@was.command(help="CSV Export")
@click.confirmation_option(prompt='\nThis is going to export all of your data into a CSV. It will take some time')
@click.option("-d", is_flag=True, help="Export most of the Plugin Data for all applications into a csv")
@click.option("-s", is_flag=True, help="Export Summary data for all applications into a csv")
def export(d, s):
    if d:
        was_detailed_export()
    if s:
        was_export()

    if not s and not d:
        click.echo("\nYou must specify a selection use -d for Detailed and -s for summary\n")
'''



import click
from .api_wrapper import tenb_connection
import datetime

tio = tenb_connection()


@click.command(help="Create or delete exclusions")
@click.option('--name', default=None, help="The name of your exclusion")
@click.option('--members', default=None, help="The members of your exclusion, IPs or subnets")
@click.option('--start', default=None, help="The start time of the exclusion - YYYY-MM-DD HH:MM")
@click.option('--end', default=None, help="The endtime of the exclusion - YYYY-MM-DD HH:MM")
@click.option('--freq', multiple=False, default=["DAILY"],
              type=click.Choice(["ONETIME", "DAILY", "WEEKLY", "MONTHLY", "YEARLY"]),
              help='The frequency of the exclusion')
@click.option('--day')
def exclude(name, members, start, end, freq, day):
    exclude_request = tio.exclusions.create(name=name, start_time=datetime.datetime.strptime(start, '%Y-%m-%d %H:%M')
                                            , end_time=datetime.datetime.strptime(end, '%Y-%m-%d %H:%M'),
                                            frequency=freq, members=list(members.split(",")), day_of_month=day)
    click.echo(exclude_request)
